package inf101.v18.rogue101.objects;

import inf101.v18.grid.GridDirection;
import inf101.v18.grid.ILocation;
import inf101.v18.rogue101.game.IGame;
import inf101.v18.rogue101.items.*;
import inf101.v18.rogue101.shared.NPC;
import javafx.scene.input.KeyCode;
import java.util.ArrayList;
import java.util.List;

public class Player implements IPlayer {
    private int hp = getMaxHealth();
    private final Backpack equipped = new Backpack();
    // True if the player wants to pick up something using the digit keys
    private boolean dropping = false;
    // True if the player wants to drop something using the digit keys
    private boolean picking = false;
    private boolean writing = false;
    private  boolean exploringChest = false;
    private String text = "";
    private String name = "Player";

    @Override
    public boolean keyPressed(IGame game, KeyCode key) {
        boolean turnConsumed = false;
        if (writing) {
            write(game, key);
            return false;
        }
        if (key == KeyCode.LEFT || key == KeyCode.A) {
            tryToMove(game, GridDirection.WEST);
            turnConsumed = true;
        } else if (key == KeyCode.RIGHT || key == KeyCode.D) {
            tryToMove(game, GridDirection.EAST);
            turnConsumed = true;
        } else if (key == KeyCode.UP || key == KeyCode.W) {
            tryToMove(game, GridDirection.NORTH);
            turnConsumed = true;
        } else if (key == KeyCode.DOWN || key == KeyCode.S) {
            tryToMove(game, GridDirection.SOUTH);
            turnConsumed = true;
        } else if (key == KeyCode.E) {
            turnConsumed = pickUpInit(game);
        } else if (key == KeyCode.Q) {
            turnConsumed = dropInit(game);
        } else if (key.isDigitKey()) {
            //Takes care of all digit keys, but we only use the first 5 for picking up and dropping.
            int keyValue = Integer.parseInt(key.getName());
            if (keyValue <= 5 && keyValue > 0) {
                if (dropping) {
                    drop(game, keyValue - 1);
                    dropping = false;
                    turnConsumed = true;
                } else if (picking) {
                    pickUp(game, keyValue - 1);
                    picking = false;
                    turnConsumed = true;
                } else if (exploringChest) {
                    loot(game, keyValue - 1);
                    exploringChest = false;
                    turnConsumed = true;
                }
            }
        } else if (key == KeyCode.N) {
            game.displayMessage("Please enter your name: ");
            writing = true;
        } else if (key == KeyCode.R) {
            IItem item = getItem(IRangedWeapon.class);
            if (item == null) {
                game.displayMessage("You do not have a ranged weapon.");
            } else {
                turnConsumed = NPC.tryAttackRanged(game, 3);
            }
        } else if (key == KeyCode.F) {
            IItem item = getItem(IMagicWeapon.class);
            if (item == null) {
                game.displayMessage("You do not have a magic weapon.");
            } else {
                turnConsumed = NPC.tryAttackRanged(game, 2);
            }
        }
        showStatus(game);
        return turnConsumed;
    }

    /**
     * Lets the user write his/her name.
     *
     * @param game  An IGame object
     * @param key   The key pressed
     */
    private void write(IGame game, KeyCode key) {
        if (key == KeyCode.BACK_SPACE) {
            text = text.substring(0, text.length() - 1);
            game.displayMessage("Please enter your name: " + text);
        } else if (key != KeyCode.ENTER) {
            text += key.toString();
            game.displayMessage("Please enter your name: " + text);
        } else {
            name = text.toLowerCase();
            text = "";
            game.displayMessage("Name set.");
            writing = false;
        }
    }

    /**
     * Initializes the picking up of an item, and does what is needed.
     *
     * @param game  An IGame object
     * @return      True if a turn was consumed. False otherwise
     */
    private boolean pickUpInit(IGame game) {
        List<IItem> items = game.getLocalItems();
        if (items.size() < 1) {
            game.displayMessage("There is nothing to pick up");
        } else {
            if (items.get(0) instanceof IStatic && items.get(0) instanceof IContainer) { //Static items are always the biggest (and hopefully the only) item on a tile.
                openChest(game, items);
            } else {
                if (items.size() == 1) {
                    pickUp(game, 0);
                    return true;
                } else {
                    StringBuilder msg = new StringBuilder("Items on this tile:");
                    for (int i = 0; i < Math.min(items.size(), 5); i++) {
                        msg.append(" [").append(i + 1).append("] ").append(firstCharToUpper(items.get(i).getName()));
                    }
                    game.displayMessage(msg.toString());
                    picking = true;
                }
            }
        }
        return false;
    }

    private void openChest(IGame game, List<IItem> items) {
        IStatic item = (IStatic)items.get(0);
        StringBuilder msg = new StringBuilder(item.getInteractMessage());
        IContainer container = (IContainer) item;
        items = container.getContent();
        for (int i = 0; i < Math.min(items.size(), 5); i++) {
            msg.append(" [").append(i + 1).append("] ").append(firstCharToUpper(items.get(i).getName()));
        }
        game.displayMessage(msg.toString());
        exploringChest = true;
    }

    /**
     * Initialized the dropping of an item, and does what is needed.
     *
     * @param game  An IGame object
     * @return      True if a turn was consumed. False otherwise
     */
    private boolean dropInit(IGame game) {
        if (equipped.getContent().size() == 1) {
            drop(game, 0);
            return true;
        } else if (equipped.getContent().size() < 1) {
            game.displayMessage("You have nothing to drop");
        } else {
            StringBuilder msg = new StringBuilder("Items to drop:");
            for (int i = 0; i < equipped.getContent().size(); i++) {
                msg.append(" [").append(i + 1).append("] ").append(firstCharToUpper(equipped.get(i).getName()));
            }
            game.displayMessage(msg.toString());
            dropping = true;
        }
        return false;
    }

    /**
     * Picks up an item at index i.
     *
     * @param game  An IGame object
     * @param i     The wanted index
     */
    private void pickUp(IGame game, int i) {
        if (!equipped.isFull()) {
            List<IItem> items = game.getLocalItems();
            if (items.size() >= i) {
                IItem pickedUp = game.pickUp(items.get(i));
                if (pickedUp != null) {
                    equipped.add(pickedUp);
                    game.displayMessage("Picked up " + pickedUp.getName());
                } else {
                    game.displayMessage("The item could not be picked up.");
                }
            } else {
                game.displayMessage("That item does not exist.");
            }
        } else {
            game.displayMessage("Your inventory is full.");
        }
    }

    /**
     * Takes an item from a chest, and gives it to the player.
     */
    private void loot(IGame game, int i) {
        if (!equipped.isFull()) {
            List<IItem> items = game.getLocalItems();
            IItem item = items.get(0);
            if (item instanceof IStatic && item instanceof IContainer) {
                IContainer container = (IContainer) item;
                IItem loot = container.getContent().get(i);
                equipped.add(loot);
                container.getContent().remove(i);
                game.displayMessage("Looted " + loot.getName());
            } else {
                game.displayMessage("It seems you are trying to loot a chest, without a chest nearby.");
            }
        } else {
            game.displayMessage("Your inventory is full.");
        }
    }

    /**
     * Drops an item at index i.
     *
     * @param game  An IGame object
     * @param i     The wanted index
     */
    private void drop(IGame game, int i) {
        if (!equipped.isEmpty() && equipped.size() > i) {
            if (game.drop(equipped.get(i))) {
                equipped.remove(i);
                game.displayMessage("Item dropped.");
            } else {
                game.displayMessage("The ground rejected the item.");
            }
        } else {
            game.displayMessage("You have nothing to drop.");
        }
    }

    /**
     * Updates the status bar with the Player's current stats.
     *
     * @param game  An IGame object
     */
    private void showStatus(IGame game) {
        List<String> items = new ArrayList<>();
        for (IItem item : equipped.getContent()) {
            String name = item.getName();
            if (name != null) {
                items.add(firstCharToUpper(name));
            }
        }
        //TODO: Add item bonuses to visible stats.
        game.formatStatus("HP: %d/%d ATK: %d DEF: %s DMG: %s INV: %s", hp, getMaxHealth(), getAttack(), getDefence(), getDamage(), String.join(",", items));
    }

    /**
     * Changes the first character in a string to uppercase.
     *
     * @param input The input string
     * @return      The input string with the first character uppercased
     */
    private String firstCharToUpper(String input) {
        if (input.length() < 1) {
            return  input;
        } else {
            return Character.toUpperCase(input.charAt(0)) + input.substring(1);
        }
    }

    private void tryToMove(IGame game, GridDirection dir) {
        ILocation loc = game.getLocation();
        if (game.canGo(dir)) {
            game.move(dir);
        } else if (loc.canGo(dir) && game.getMap().hasActors(loc.go(dir))) {
            NPC.playSound("audio/Sword Swing-SoundBible.com-639083727.wav");
            game.attack(dir, game.getMap().getActors(loc.go(dir)).get(0));
        } else {
           game.displayMessage("Umm, it is not possible to move in that direction.");
        }
    }

    @Override
    public int getAttack() {
        return 50;
    }

    @Override
    public int getDamage() {
        return 20;
    }

    @Override
    public IItem getItem(Class<?> type) {
        for (IItem item : equipped.getContent()) {
            if (type.isInstance(item)) {
                return item;
            }
        }
        return null;
    }

    @Override
    public int getCurrentHealth() {
        return hp;
    }

    @Override
    public int getDefence() {
        return 30;
    }

    @Override
    public int getMaxHealth() {
        return 1000;
    }

    @Override
    public String getName() {
        return firstCharToUpper(name);
    }

    @Override
    public int getSize() {
        return 10;
    }

    @Override
    public String getPrintSymbol() {
        return  "\u001b[96m" + "@" + "\u001b[0m";
    }

    @Override
    public String getSymbol() {
        return "@";
    }

    @Override
    public int handleDamage(IGame game, IItem source, int amount) {
        hp -= amount;
        showStatus(game);
        if (hp < 1) {
            game.displayStatus("Game Over");
        }
        return amount;
    }

    @Override
    public int getVision() {
        //TODO: Increase vision based on equipped items
        return 3;
    }
}
